@@ -1,124 +1,127 @@
+from __future__ import annotations
+
 import collections
 import contextlib
 import functools
 import inspect
 import json as pyjson
-import logging
 import os
 import re
 import typing
 from collections.abc import Mapping
-from enum import Enum
 from pathlib import Path
-from urllib.parse import ParseResult, urlparse
 
 import marshmallow as ma
 from dotenv.main import _walk_to_root, load_dotenv
 
-__all__ = ["EnvError", "Env"]
+from . import fields
+from .exceptions import (
+    EnvError,
+    EnvSealedError,
+    EnvValidationError,
+    ParserConflictError,
+)
+
+if typing.TYPE_CHECKING:
+    import datetime as dt
+    import decimal
+    import uuid
+    from urllib.parse import ParseResult
+
+    from .types import (
+        DictFieldMethod,
+        EnumFieldMethod,
+        ErrorMapping,
+        FieldFactory,
+        FieldMethod,
+        ListFieldMethod,
+        ParserMethod,
+        Subcast,
+        TimeDeltaFieldMethod,
+    )
+
+    try:
+        from dj_database_url import DBConfig
+    except ImportError:
+        pass
+
+__all__ = ["Env", "EnvError", "FileAwareEnv", "ValidationError", "env"]
 
 _T = typing.TypeVar("_T")
 _StrType = str
 _BoolType = bool
-
-ErrorMapping = typing.Mapping[str, typing.List[str]]
-ErrorList = typing.List[str]
-FieldFactory = typing.Callable[..., ma.fields.Field]
-Subcast = typing.Union[typing.Type, typing.Callable[..., _T], ma.fields.Field]
-FieldType = typing.Type[ma.fields.Field]
-FieldOrFactory = typing.Union[FieldType, FieldFactory]
-ParserMethod = typing.Callable
-
+_IntType = int
+_ListType = list
+_DictType = dict
 
 _EXPANDED_VAR_PATTERN = re.compile(r"(?<!\\)\$\{([A-Za-z0-9_]+)(:-[^\}:]*)?\}")
 
 
-class EnvError(ValueError):
-    """Raised when an environment variable or
-    if a required environment variable is unset.
-    """
-
-
-class EnvValidationError(EnvError):
-    def __init__(
-        self, message: str, error_messages: typing.Union[ErrorList, ErrorMapping]
-    ):
-        self.error_messages = error_messages
-        super().__init__(message)
-
-
-class EnvSealedError(TypeError, EnvError):
-    pass
-
-
-class ParserConflictError(ValueError):
-    """Raised when adding a custom parser that conflicts
-    with a built-in parser method.
-    """
+# Reexport marshmallow's validate module and ValidationError.
+validate = ma.validate
+# Custom validators should raise this for invalid input.
+ValidationError = ma.ValidationError
 
 
 def _field2method(
-    field_or_factory: FieldOrFactory,
+    field_or_factory: type[ma.fields.Field] | FieldFactory,
     method_name: str,
     *,
-    preprocess: typing.Optional[typing.Callable] = None,
-    preprocess_kwarg_names: typing.Sequence[str] = tuple(),
-) -> ParserMethod:
+    preprocess: typing.Callable | None = None,
+    preprocess_kwarg_names: typing.Sequence[str] = (),
+) -> typing.Any:
     def method(
-        self: "Env",
+        self: Env,
         name: str,
-        default: typing.Any = ma.missing,
-        subcast: typing.Optional[Subcast] = None,
+        default: typing.Any = ...,
+        subcast: Subcast[_T] | None = None,
         *,
         # Subset of relevant marshmallow.Field kwargs
-        load_default: typing.Any = ma.missing,
-        validate: typing.Optional[
-            typing.Union[
-                typing.Callable[[typing.Any], typing.Any],
-                typing.Iterable[typing.Callable[[typing.Any], typing.Any]],
-            ]
-        ] = None,
-        required: bool = False,
-        allow_none: typing.Optional[bool] = None,
-        error_messages: typing.Optional[typing.Dict[str, str]] = None,
-        metadata: typing.Optional[typing.Mapping[str, typing.Any]] = None,
+        validate: (
+            typing.Callable[[typing.Any], typing.Any]
+            | typing.Iterable[typing.Callable[[typing.Any], typing.Any]]
+            | None
+        ) = None,
+        # Additional kwargs are passed to Field constructor
         **kwargs,
-    ) -> typing.Optional[_T]:
+    ) -> _T | None:
         if self._sealed:
             raise EnvSealedError(
-                "Env has already been sealed. New values cannot be parsed."
+                "Env has already been sealed. New values cannot be parsed.",
             )
-        field_kwargs = dict(
-            validate=validate,
-            required=required,
-            allow_none=allow_none,
-            load_default=load_default or default,
-            error_messages=error_messages,
-            metadata=metadata,
-        )
+        load_default = default if default is not Ellipsis else ma.missing
         preprocess_kwargs = {
             name: kwargs.pop(name) for name in preprocess_kwarg_names if name in kwargs
         }
         if isinstance(field_or_factory, type) and issubclass(
-            field_or_factory, ma.fields.Field
+            field_or_factory,
+            ma.fields.Field,
         ):
-            field = field_or_factory(**field_kwargs, **kwargs)
+            field = field_or_factory(
+                validate=validate,
+                load_default=load_default,
+                **kwargs,
+            )
         else:
-            parsed_subcast = _make_subcast_field(subcast)
-            field = field_or_factory(subcast=parsed_subcast, **field_kwargs)
-        parsed_key, value, proxied_key = self._get_from_environ(
-            name, field.load_default
-        )
+            parsed_subcast = _make_subcast_field(subcast) if subcast else ma.fields.Raw
+            field = typing.cast("FieldFactory", field_or_factory)(
+                subcast=parsed_subcast,
+                validate=validate,
+                load_default=load_default,
+            )
+        parsed_key, value, proxied_key = self._get_from_environ(name, default=Ellipsis)
         self._fields[parsed_key] = field
         source_key = proxied_key or parsed_key
-        if value is ma.missing:
+        if value is Ellipsis:
+            if default is not Ellipsis:
+                self._values[parsed_key] = default
+                return default
             if self.eager:
                 raise EnvError(
-                    f'Environment variable "{proxied_key or parsed_key}" not set'
+                    f'Environment variable "{proxied_key or parsed_key}" not set',
                 )
-            else:
-                self._errors[parsed_key].append("Environment variable not set.")
-                return None
+            self._errors[parsed_key].append("Environment variable not set.")
+            return None
         try:
             if preprocess:
                 value = preprocess(value, **preprocess_kwargs)
@@ -132,35 +135,34 @@ def _field2method(
             self._errors[parsed_key].extend(error.messages)
         else:
             self._values[parsed_key] = value
-        return value
+        return typing.cast("_T | None", value)
 
     method.__name__ = method_name
     return method
 
 
-def _func2method(func: typing.Callable, method_name: str) -> ParserMethod:
+def _func2method(func: typing.Callable[..., _T], method_name: str) -> typing.Any:
     def method(
-        self: "Env",
+        self: Env,
         name: str,
-        default: typing.Any = ma.missing,
+        default: typing.Any = ...,
         **kwargs,
-    ) -> typing.Optional[_T]:
+    ) -> _T | None:
         if self._sealed:
             raise EnvSealedError(
-                "Env has already been sealed. New values cannot be parsed."
+                "Env has already been sealed. New values cannot be parsed.",
             )
         parsed_key, raw_value, proxied_key = self._get_from_environ(name, default)
-        self._fields[parsed_key] = ma.fields.Field()
+        self._fields[parsed_key] = ma.fields.Raw()
         source_key = proxied_key or parsed_key
-        if raw_value is ma.missing:
+        if raw_value is Ellipsis:
             if self.eager:
                 raise EnvError(
-                    f'Environment variable "{proxied_key or parsed_key}" not set'
+                    f'Environment variable "{proxied_key or parsed_key}" not set',
                 )
-            else:
-                self._errors[parsed_key].append("Environment variable not set.")
-                return None
-        if raw_value or raw_value == "":
+            self._errors[parsed_key].append("Environment variable not set.")
+            return None
+        if raw_value or raw_value == "":  # noqa: SIM108
             value = raw_value
         else:
             value = None
@@ -180,15 +182,15 @@ def _func2method(func: typing.Callable, method_name: str) -> ParserMethod:
             self._errors[parsed_key].extend(messages)
         else:
             self._values[parsed_key] = value
-        return value
+        return typing.cast("_T | None", value)
 
     method.__name__ = method_name
     return method
 
 
 def _make_subcast_field(
-    subcast: typing.Optional[Subcast],
-) -> typing.Type[ma.fields.Field]:
+    subcast: Subcast,
+) -> type[ma.fields.Field]:
     if isinstance(subcast, type) and subcast in ma.Schema.TYPE_MAPPING:
         inner_field = ma.Schema.TYPE_MAPPING[subcast]
     elif isinstance(subcast, type) and issubclass(subcast, ma.fields.Field):
@@ -197,89 +199,80 @@ def _make_subcast_field(
 
         class SubcastField(ma.fields.Field):
             def _deserialize(self, value, *args, **kwargs):
-                func = typing.cast(typing.Callable[..., _T], subcast)
-                return func(value)
+                return subcast(value)
 
         inner_field = SubcastField
     else:
-        inner_field = ma.fields.Field
+        inner_field = ma.fields.Raw
     return inner_field
 
 
-def _make_list_field(*, subcast: typing.Optional[type], **kwargs) -> ma.fields.List:
-    inner_field = _make_subcast_field(subcast)
+def _make_list_field(*, subcast: Subcast | None, **kwargs) -> ma.fields.List:
+    inner_field = _make_subcast_field(subcast) if subcast else ma.fields.Raw
     return ma.fields.List(inner_field, **kwargs)
 
 
 def _preprocess_list(
-    value: typing.Union[str, typing.Iterable], *, delimiter: str = ",", **kwargs
+    value: str | typing.Iterable,
+    *,
+    delimiter: str = ",",
+    **kwargs,
 ) -> typing.Iterable:
     if ma.utils.is_iterable_but_not_string(value) or value is None:
         return value
-    return typing.cast(str, value).split(delimiter) if value != "" else []
+    return typing.cast("str", value).split(delimiter) if value != "" else []
 
 
 def _preprocess_dict(
-    value: typing.Union[str, typing.Mapping],
+    value: str | typing.Mapping,
     *,
-    subcast_keys: typing.Optional[Subcast] = None,
-    subcast_values: typing.Optional[Subcast] = None,
+    subcast_keys: Subcast | None = None,
+    subcast_values: Subcast | None = None,
     delimiter: str = ",",
+    key_value_delimiter: str = "=",
     **kwargs,
 ) -> typing.Mapping:
     if isinstance(value, Mapping):
         return value
-    subcast_keys_instance: ma.fields.Field = _make_subcast_field(subcast_keys)(**kwargs)
-    subcast_values_instance: ma.fields.Field = _make_subcast_field(subcast_values)(
-        **kwargs
-    )
+    subcast_keys_instance: ma.fields.Field
+    if subcast_keys:
+        subcast_keys_instance = _make_subcast_field(subcast_keys)(**kwargs)
+    else:
+        subcast_keys_instance = ma.fields.Raw()
+    subcast_values_instance: ma.fields.Field
+    if subcast_values:
+        subcast_values_instance = _make_subcast_field(subcast_values)(**kwargs)
+    else:
+        subcast_values_instance = ma.fields.Raw()
 
     return {
         subcast_keys_instance.deserialize(
-            key.strip()
+            key.strip(),
         ): subcast_values_instance.deserialize(val.strip())
-        for key, val in (item.split("=", 1) for item in value.split(delimiter) if value)
+        for key, val in (
+            item.split(key_value_delimiter, 1) for item in value.split(delimiter) if value
+        )
     }
 
 
-def _preprocess_json(value: typing.Union[str, typing.Mapping, typing.List], **kwargs):
+def _preprocess_json(value: str | typing.Mapping | list, **kwargs):
     try:
         if isinstance(value, str):
             return pyjson.loads(value)
-        elif isinstance(value, dict) or isinstance(value, list) or value is None:
+        if isinstance(value, (dict, list)) or value is None:
             return value
-        else:
-            raise ma.ValidationError("Not valid JSON.")
+        raise ma.ValidationError("Not valid JSON.")
     except pyjson.JSONDecodeError as error:
         raise ma.ValidationError("Not valid JSON.") from error
 
 
-_EnumT = typing.TypeVar("_EnumT", bound=Enum)
-
-
-def _enum_parser(value, type: typing.Type[_EnumT], ignore_case: bool = False) -> _EnumT:
-    invalid_exc = ma.ValidationError(f"Not a valid '{type.__name__}' enum.")
-
-    if not ignore_case:
-        try:
-            return type[value]
-        except KeyError as error:
-            raise invalid_exc from error
-
-    for enum_value in type:
-        if enum_value.name.lower() == value.lower():
-            return enum_value
-
-    raise invalid_exc
-
-
-def _dj_db_url_parser(value: str, **kwargs) -> dict:
+def _dj_db_url_parser(value: str, **kwargs) -> DBConfig:
     try:
-        import dj_database_url
+        import dj_database_url  # noqa: PLC0415
     except ImportError as error:
         raise RuntimeError(
             "The dj_db_url parser requires the dj-database-url package. "
-            "You can install it with: pip install dj-database-url"
+            "You can install it with: pip install dj-database-url",
         ) from error
     try:
         return dj_database_url.parse(value, **kwargs)
@@ -289,11 +282,11 @@ def _dj_db_url_parser(value: str, **kwargs) -> dict:
 
 def _dj_email_url_parser(value: str, **kwargs) -> dict:
     try:
-        import dj_email_url
+        import dj_email_url  # noqa: PLC0415
     except ImportError as error:
         raise RuntimeError(
             "The dj_email_url parser requires the dj-email-url package. "
-            "You can install it with: pip install dj-email-url"
+            "You can install it with: pip install dj-email-url",
         ) from error
     try:
         return dj_email_url.parse(value, **kwargs)
@@ -303,11 +296,11 @@ def _dj_email_url_parser(value: str, **kwargs) -> dict:
 
 def _dj_cache_url_parser(value: str, **kwargs) -> dict:
     try:
-        import django_cache_url
+        import django_cache_url  # noqa: PLC0415
     except ImportError as error:
         raise RuntimeError(
             "The dj_cache_url parser requires the django-cache-url package. "
-            "You can install it with: pip install django-cache-url"
+            "You can install it with: pip install django-cache-url",
         ) from error
     try:
         return django_cache_url.parse(value, **kwargs)
@@ -317,105 +310,82 @@ def _dj_cache_url_parser(value: str, **kwargs) -> dict:
         raise ma.ValidationError(error.args[0]) from error
 
 
-class URLField(ma.fields.URL):
-    def _serialize(self, value: ParseResult, *args, **kwargs) -> str:
-        return value.geturl()
-
-    # Override deserialize rather than _deserialize because we need
-    # to call urlparse *after* validation has occurred
-    def deserialize(
-        self,
-        value: str,
-        attr: typing.Optional[str] = None,
-        data: typing.Optional[typing.Mapping] = None,
-        **kwargs,
-    ) -> ParseResult:
-        ret = super().deserialize(value, attr, data, **kwargs)
-        return urlparse(ret)
-
-
-class PathField(ma.fields.Str):
-    def _deserialize(self, value, *args, **kwargs) -> Path:
-        if isinstance(value, Path):
-            return value
-        ret = super()._deserialize(value, *args, **kwargs)
-        return Path(ret)
-
-
-class LogLevelField(ma.fields.Int):
-    def _format_num(self, value) -> int:
-        try:
-            return super()._format_num(value)
-        except (TypeError, ValueError) as error:
-            value = value.upper()
-            if hasattr(logging, value) and isinstance(getattr(logging, value), int):
-                return getattr(logging, value)
-            else:
-                raise ma.ValidationError("Not a valid log level.") from error
-
-
 class Env:
     """An environment variable reader."""
 
-    __call__: ParserMethod = _field2method(ma.fields.Field, "__call__")
+    __call__: FieldMethod[typing.Any] = _field2method(ma.fields.Raw, "__call__")
 
-    int = _field2method(ma.fields.Int, "int")
-    bool = _field2method(ma.fields.Bool, "bool")
-    str = _field2method(ma.fields.Str, "str")
-    float = _field2method(ma.fields.Float, "float")
-    decimal = _field2method(ma.fields.Decimal, "decimal")
-    list = _field2method(
+    int: FieldMethod[int] = _field2method(ma.fields.Int, "int")
+    bool: FieldMethod[bool] = _field2method(ma.fields.Bool, "bool")
+    str: FieldMethod[str] = _field2method(ma.fields.Str, "str")
+    float: FieldMethod[float] = _field2method(ma.fields.Float, "float")
+    decimal: FieldMethod[decimal.Decimal] = _field2method(ma.fields.Decimal, "decimal")
+    list: ListFieldMethod = _field2method(
         _make_list_field,
         "list",
         preprocess=_preprocess_list,
         preprocess_kwarg_names=("subcast", "delimiter"),
     )
-    dict = _field2method(
+    dict: DictFieldMethod = _field2method(
         ma.fields.Dict,
         "dict",
         preprocess=_preprocess_dict,
         preprocess_kwarg_names=(
-            "subcast",
             "subcast_keys",
             "subcast_key",
             "subcast_values",
             "delimiter",
+            "key_value_delimiter",
         ),
     )
-    json = _field2method(ma.fields.Field, "json", preprocess=_preprocess_json)
-    datetime = _field2method(ma.fields.DateTime, "datetime")
-    date = _field2method(ma.fields.Date, "date")
-    time = _field2method(ma.fields.Time, "time")
-    path = _field2method(PathField, "path")
-    log_level = _field2method(LogLevelField, "log_level")
-    timedelta = _field2method(ma.fields.TimeDelta, "timedelta")
-    uuid = _field2method(ma.fields.UUID, "uuid")
-    url = _field2method(URLField, "url")
-    enum = _func2method(_enum_parser, "enum")
+    json: FieldMethod[_ListType | _DictType] = _field2method(
+        ma.fields.Raw,
+        "json",
+        preprocess=_preprocess_json,
+    )
+    datetime: FieldMethod[dt.datetime] = _field2method(ma.fields.DateTime, "datetime")
+    date: FieldMethod[dt.date] = _field2method(ma.fields.Date, "date")
+    time: FieldMethod[dt.time] = _field2method(ma.fields.Time, "time")
+    timedelta: TimeDeltaFieldMethod = _field2method(fields.TimeDelta, "timedelta")
+    path: FieldMethod[Path] = _field2method(fields.Path, "path")
+    log_level: FieldMethod[_IntType] = _field2method(fields.LogLevel, "log_level")
+
+    uuid: FieldMethod[uuid.UUID] = _field2method(ma.fields.UUID, "uuid")
+    url: FieldMethod[ParseResult] = _field2method(fields.Url, "url")
+
+    enum: EnumFieldMethod = _field2method(ma.fields.Enum, "enum")
     dj_db_url = _func2method(_dj_db_url_parser, "dj_db_url")
     dj_email_url = _func2method(_dj_email_url_parser, "dj_email_url")
     dj_cache_url = _func2method(_dj_cache_url_parser, "dj_cache_url")
 
-    def __init__(self, *, eager: _BoolType = True, expand_vars: _BoolType = False):
+    def __init__(
+        self,
+        *,
+        eager: _BoolType = True,
+        expand_vars: _BoolType = False,
+        prefix: _StrType | None = None,
+    ):
         self.eager = eager
         self._sealed: bool = False
         self.expand_vars = expand_vars
-        self._fields: typing.Dict[_StrType, typing.Union[ma.fields.Field, type]] = {}
-        self._values: typing.Dict[_StrType, typing.Any] = {}
+        self._fields: dict[_StrType, ma.fields.Field] = {}
+        self._values: dict[_StrType, typing.Any] = {}
         self._errors: ErrorMapping = collections.defaultdict(list)
-        self._prefix: typing.Optional[_StrType] = None
-        self.__custom_parsers__: typing.Dict[_StrType, ParserMethod] = {}
+        self._prefix: _StrType | None = prefix
+        self.__custom_parsers__: dict[_StrType, ParserMethod] = {}
 
     def __repr__(self) -> _StrType:
-        return f"<{self.__class__.__name__}(eager={self.eager}, expand_vars={self.expand_vars})>"  # noqa: E501
+        return f"<{self.__class__.__name__}(eager={self.eager}, expand_vars={self.expand_vars})>"
 
     @staticmethod
     def read_env(
-        path: typing.Optional[_StrType] = None,
+        path: _StrType | Path | None = None,
+        *,
         recurse: _BoolType = True,
         verbose: _BoolType = False,
         override: _BoolType = False,
-    ) -> _BoolType:
+        return_path: _BoolType = False,
+    ) -> _BoolType | _StrType | None:
         """Read a .env file into os.environ.
 
         If .env is not found in the directory from which this method is called,
@@ -423,13 +393,15 @@ class Env:
         file is found. If you do not wish to recurse up the tree, you may pass
         False as a second positional argument.
         """
+        env_path = None
+        is_env_loaded = False
         if path is None:
             # By default, start search from the same directory this function is called
             current_frame = inspect.currentframe()
             if current_frame is None:
                 raise RuntimeError("Could not get current call frame.")
             frame = current_frame.f_back
-            assert frame is not None
+            assert frame is not None  # noqa: S101
             caller_dir = Path(frame.f_code.co_filename).parent.resolve()
             start = caller_dir / ".env"
         else:
@@ -443,13 +415,24 @@ class Env:
             for dirname in _walk_to_root(start_dir):
                 check_path = Path(dirname) / env_name
                 if check_path.exists():
-                    return load_dotenv(check_path, verbose=verbose, override=override)
-            return False
+                    is_env_loaded = load_dotenv(
+                        check_path,
+                        verbose=verbose,
+                        override=override,
+                    )
+                    env_path = str(check_path)
+                    break
+
         else:
-            return load_dotenv(str(start), verbose=verbose, override=override)
+            is_env_loaded = load_dotenv(str(start), verbose=verbose, override=override)
+            env_path = str(start)
+
+        if return_path:
+            return env_path
+        return is_env_loaded
 
     @contextlib.contextmanager
-    def prefixed(self, prefix: _StrType) -> typing.Iterator["Env"]:
+    def prefixed(self, prefix: _StrType) -> typing.Iterator[Env]:
         """Context manager for parsing envvars with a common prefix."""
         try:
             old_prefix = self._prefix
@@ -473,7 +456,8 @@ class Env:
             error_messages = dict(self._errors)
             self._errors = {}
             raise EnvValidationError(
-                f"Environment variables invalid: {error_messages}", error_messages
+                f"Environment variables invalid: {error_messages}",
+                error_messages,
             )
 
     def __getattr__(self, name: _StrType):
@@ -488,13 +472,13 @@ class Env:
         """
         if hasattr(self, name):
             raise ParserConflictError(
-                f"Env already has a method with name '{name}'. Use a different name."
+                f"Env already has a method with name '{name}'. Use a different name.",
             )
         self.__custom_parsers__[name] = _func2method(func, method_name=name)
-        return None
 
     def parser_for(
-        self, name: _StrType
+        self,
+        name: _StrType,
     ) -> typing.Callable[[typing.Callable], typing.Callable]:
         """Decorator that registers a new parser method with the name ``name``.
         The decorated function must receive the input value for an environment variable.
@@ -506,9 +490,7 @@ class Env:
 
         return decorator
 
-    def add_parser_from_field(
-        self, name: _StrType, field_cls: typing.Type[ma.fields.Field]
-    ):
+    def add_parser_from_field(self, name: _StrType, field_cls: type[ma.fields.Field]):
         """Register a new parser method with name ``name``,
         given a marshmallow ``Field``.
         """
@@ -522,8 +504,12 @@ class Env:
         return schema.dump(self._values)
 
     def _get_from_environ(
-        self, key: _StrType, default: typing.Any, *, proxied: _BoolType = False
-    ) -> typing.Tuple[_StrType, typing.Any, typing.Optional[_StrType]]:
+        self,
+        key: _StrType,
+        default: typing.Any,
+        *,
+        proxied: _BoolType = False,
+    ) -> tuple[_StrType, typing.Any, _StrType | None]:
         """Access a value from os.environ. Handles proxied variables,
         e.g. SMTP_LOGIN={{MAILGUN_LOGIN}}.
 
@@ -535,18 +521,18 @@ class Env:
         to get a proxy env key.
         """
         env_key = self._get_key(key, omit_prefix=proxied)
-        value = os.environ.get(env_key, default)
+        value = self._get_value(env_key, default)
         if hasattr(value, "strip"):
-            expand_match = self.expand_vars and _EXPANDED_VAR_PATTERN.match(value)
+            expand_match = self.expand_vars and _EXPANDED_VAR_PATTERN.fullmatch(value)
             if expand_match:  # Full match expand_vars - special case keep default
                 proxied_key: _StrType = expand_match.groups()[0]
-                subs_default: typing.Optional[_StrType] = expand_match.groups()[1]
+                subs_default: _StrType | None = expand_match.groups()[1]
                 if subs_default is not None:
                     default = subs_default[2:]
                 elif (
                     value == default
                 ):  # if we have used default, don't use it recursively
-                    default = ma.missing
+                    default = Ellipsis
                 return (
                     key,
                     self._get_from_environ(proxied_key, default, proxied=True)[1],
@@ -568,12 +554,12 @@ class Env:
         for match in _EXPANDED_VAR_PATTERN.finditer(value):
             env_key = match.group(1)
             env_default = match.group(2)
-            if env_default is None:
-                env_default = ma.missing
+            if env_default is None:  # noqa: SIM108
+                env_default = Ellipsis
             else:
                 env_default = env_default[2:]  # trim ':-' from default
             _, env_value, _ = self._get_from_environ(env_key, env_default, proxied=True)
-            if env_value is ma.missing:
+            if env_value is Ellipsis:
                 return parsed_key, env_value, env_key
             ret += value[prev_start : match.start()] + env_value
             prev_start = match.end()
@@ -583,3 +569,40 @@ class Env:
 
     def _get_key(self, key: _StrType, *, omit_prefix: _BoolType = False) -> _StrType:
         return self._prefix + key if self._prefix and not omit_prefix else key
+
+    def _get_value(self, env_key: _StrType, default: typing.Any) -> typing.Any:
+        return os.environ.get(env_key, default)
+
+
+class FileAwareEnv(Env):
+    """An environment variable reader that supports reading values from files."""
+
+    def __init__(
+        self,
+        *,
+        file_suffix: _StrType = "_FILE",
+        eager: _BoolType = True,
+        expand_vars: _BoolType = False,
+        prefix: _StrType | None = None,
+        strip_whitespace: _BoolType = False,
+    ):
+        self.file_suffix: _StrType = file_suffix
+        self.strip_whitespace: _BoolType = strip_whitespace
+        super().__init__(eager=eager, expand_vars=expand_vars, prefix=prefix)
+
+    def _get_value(self, env_key: _StrType, default: typing.Any) -> typing.Any:
+        """Return the contents of the file referenced in key <env_key>_FILE, if present."""
+        file_key = f"{env_key}{self.file_suffix}"
+        if file_path := os.environ.get(file_key, None):
+            try:
+                value: _StrType = Path(file_path).read_text()
+                return value.strip() if self.strip_whitespace else value
+            except (FileNotFoundError, IsADirectoryError, PermissionError) as err:
+                raise ValueError(
+                    f"The value of {file_key} must be a readable file path."
+                ) from err
+        return super()._get_value(env_key, default)
+
+
+# Singleton instance, for convenience
+env = Env()
